@@ -31,7 +31,10 @@ import ProxyPatternPool as ppp  # type: ignore
 from flask import (
     Response, Request, request, session, jsonify, Blueprint, make_response,
     abort, redirect, url_for, after_this_request, send_file, current_app, g,
-    send_from_directory, escape, Markup, render_template)
+    send_from_directory, escape, Markup, render_template, get_flashed_messages,
+    has_app_context, has_request_context, render_template_string, stream_template,
+    stream_template_string, stream_with_context,
+)
 
 import logging
 log = logging.getLogger("fsa")
@@ -69,7 +72,7 @@ class string(str):
 
 # "JsonData = json.loads" would do:-)
 class JsonData:
-    """Magic JSON Type."""
+    """Magic JSON type."""
     pass
 
 
@@ -144,23 +147,35 @@ class Flask(flask.Flask):
     - several additional methods are provided: `get_user_pass`,
       `user_in_group`, `check_password`, `hash_password`, `create_token`,
       `get_user`, `current_user`, `clear_caches`, `cast`, `object_perms`,
-      `user_oauth`, `password_quality` and `password_check`.
+      `user_scope`, `password_quality`, `password_check`, `add_group`,
+      `add_scope`.
     """
 
-    def __init__(self, *args, debug: Optional[bool] = None, **kwargs):
+    def __init__(self, *args, debug: bool = False, **kwargs):
+        # extract FSA-specific directives
+        fsaconf: Dict[str, Any] = {}
+        for key, val in kwargs.items():
+            if key.startswith("FSA_"):
+                fsaconf[key] = val
+        for key in fsaconf:
+            del kwargs[key]
+        # Flask initialization
         super().__init__(*args, **kwargs)
-        self._fsa = FlaskSimpleAuth(self, debug=debug)
+        # FSA initialization
+        self._fsa = FlaskSimpleAuth(self, debug=debug, **fsaconf)
         # overwritten late because called by upper Flask initialization for "static"
         setattr(self, "add_url_rule", self._fsa.add_url_rule)
         # forward hooks
         self.get_user_pass = self._fsa.get_user_pass
         self.user_in_group = self._fsa.user_in_group
         self.object_perms = self._fsa.object_perms
-        self.user_oauth = self._fsa.user_oauth
+        self.user_scope = self._fsa.user_scope
         self.cast = self._fsa.cast
         self.special_parameter = self._fsa.special_parameter
         self.password_quality = self._fsa.password_quality
         self.password_check = self._fsa.password_check
+        self.add_group = self._fsa.add_group
+        self.add_scope = self._fsa.add_scope
         # forward methods
         self.check_password = self._fsa.check_password
         self.hash_password = self._fsa.hash_password
@@ -182,8 +197,6 @@ _DIRECTIVES = {
     "FSA_DEBUG", "FSA_LOGGING_LEVEL",
     # general settings
     "FSA_SECURE", "FSA_SERVER_ERROR", "FSA_NOT_FOUND_ERROR", "FSA_LOCAL",
-    # parameter handing
-    "FSA_REJECT_UNEXPECTED_PARAM",
     # register hooks
     "FSA_GET_USER_PASS", "FSA_USER_IN_GROUP", "FSA_CAST",
     "FSA_OBJECT_PERMS", "FSA_SPECIAL_PARAMETER",
@@ -196,6 +209,10 @@ _DIRECTIVES = {
     "FSA_PASSWORD_SCHEME", "FSA_PASSWORD_OPTS", "FSA_PASSWORD_CHECK",
     "FSA_PASSWORD_LEN", "FSA_PASSWORD_RE", "FSA_PASSWORD_QUALITY",
     "FSA_HTTP_AUTH_OPTS",
+    # authorization
+    "FSA_AUTHZ_GROUPS", "FSA_AUTHZ_SCOPES",
+    # parameter handing
+    "FSA_REJECT_UNEXPECTED_PARAM",
     # internal caching
     "FSA_CACHE", "FSA_CACHE_SIZE", "FSA_CACHE_OPTS", "FSA_CACHE_PREFIX",
     # web-oriented settings
@@ -217,13 +234,14 @@ _DEFAULT_PASSWORD_OPTS = {"bcrypt__default_rounds": 4, "bcrypt__default_ident":
 class FlaskSimpleAuth:
     """Flask extension for authentication, authorization and parameters."""
 
-    def __init__(self, app: Optional[flask.Flask] = None, debug: Optional[bool] = False):
+    def __init__(self, app: flask.Flask, debug: bool = False, **config):
         """Constructor parameter: flask application to extend."""
         self._debug = debug
         if debug:
             logging.warning("FlaskSimpleAuth running in debug mode")
             log.setLevel(logging.DEBUG)
         self._app = app
+        self._app.config.update(**config)
         # hooks
         self._get_user_pass = None
         self._user_in_group = None
@@ -257,6 +275,8 @@ class FlaskSimpleAuth:
         self._not_found_error: int = _DEFAULT_NOT_FOUND_ERROR
         self._secure: bool = True
         self._names: Set[str] = set()
+        self._groups: Set[Union[str, int]] = set()
+        self._scopes: Set[str] = set()
         self._local: Any = None
         # actual main initialization is deferred to `init_app`
         self._initialized = False
@@ -443,10 +463,20 @@ class FlaskSimpleAuth:
         assert domain in self._object_perms
         return self._object_perms[domain](user, oid, mode)
 
-    def user_oauth(self, scope):
+    def user_scope(self, scope):
         """Is scope in the current user scope."""
         return self._local.scopes and scope in self._local.scopes
 
+    def add_group(self, *groups):
+        """Add some groups."""
+        for grp in groups:
+            self._groups.add(grp)
+
+    def add_scope(self, *scopes):
+        """Add some scopes."""
+        for scope in scopes:
+            self._scopes.add(scope)
+
     #
     # DEFERRED INITIALIZATIONS
     #
@@ -491,7 +521,6 @@ class FlaskSimpleAuth:
         else:
             raise self._Bad(f"unexpected FSA_LOCAL value: {local}")
         self._local = Local()
-        # NOTE threading.local might not be enough?
         # whether to only allow secure requests
         self._secure = conf.get("FSA_SECURE", True)
         # status code for some errors errors
@@ -517,6 +546,11 @@ class FlaskSimpleAuth:
                 raise self._Bad(f"unexpected auth: {a}")
         self._local.auth = self._auth  # type: ignore
         #
+        # authorize
+        #
+        self._groups.update(conf.get("FSA_AUTHZ_GROUPS", []))
+        self._scopes.update(conf.get("FSA_AUTHZ_SCOPES", []))
+        #
         # web apps…
         #
         self._cors: bool = conf.get("FSA_CORS", False)
@@ -1226,11 +1260,11 @@ class FlaskSimpleAuth:
     # INTERNAL DECORATORS
     #
     # _authenticate: set self._user
-    #   _oauth_auth: check OAuth scope authorization
-    #   _group_auth: check group authorization
+    #  _oauth_authz: check OAuth scope authorization
+    #  _group_authz: check group authorization
+    #     _no_authz: validate that no authorization was needed
     #   _parameters: handle HTTP/JSON to python parameter translation
-    #    _perm_auth: check per-object permissions
-    #   _any_noauth: validate that no authorization was needed
+    #   _perm_authz: check per-object permissions
     #
     def _safe_call(self, path, level, fun, *args, **kwargs):
         """Call a route function ensuring a response whatever."""
@@ -1278,9 +1312,14 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    def _oauth_auth(self, path, *scopes):
+    def _oauth_authz(self, path, *scopes):
         """Decorator to authorize OAuth scopes (token-provided authz)."""
 
+        if self._scopes:
+            for scope in scopes:
+                if scope not in self._scopes:
+                    raise self._Bad(f"unexpected scope {scope}")
+
         def decorate(fun: Callable):
 
             @functools.wraps(fun)
@@ -1289,7 +1328,7 @@ class FlaskSimpleAuth:
                 self._local.need_authorization = False
 
                 for scope in scopes:
-                    if not self.user_oauth(scope):
+                    if not self.user_scope(scope):
                         return self._Res("", 403)
 
                 return self._safe_call(path, "oauth authorization", fun, *args, **kwargs)
@@ -1298,12 +1337,17 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    def _group_auth(self, path, *groups):
+    def _group_authz(self, path, *groups):
         """Decorator to authorize user groups."""
 
-        for group in _PREDEFS:
-            if group in groups:
-                raise self._Bad(f"unexpected predefined {group}")
+        for grp in _PREDEFS:
+            if grp in groups:
+                raise self._Bad(f"unexpected predefined {grp}")
+
+        if self._groups:
+            for grp in groups:
+                if grp not in self._groups:
+                    raise self._Bad(f"unexpected group {grp}")
 
         if not self._user_in_group:  # pragma: no cover
             raise self._Bad(f"user_in_group callback needed for group authorization on {path}")
@@ -1317,9 +1361,9 @@ class FlaskSimpleAuth:
                 self._local.need_authorization = False
 
                 # check against all authorized groups/roles
-                for group in groups:
+                for grp in groups:
                     try:
-                        ok = self._user_in_group(self._local.user, group)
+                        ok = self._user_in_group(self._local.user, grp)
                     except ErrorResponse as e:
                         return self._Res(e.message, e.status)
                     except Exception as e:
@@ -1338,6 +1382,20 @@ class FlaskSimpleAuth:
 
         return decorate
 
+    # just to record that no authorization check was needed
+    def _no_authz(self, path, *groups):
+
+        def decorate(fun: Callable):
+
+            @functools.wraps(fun)
+            def wrapper(*args, **kwargs):
+                self._local.need_authorization = False
+                return self._safe_call(path, "no authorization", fun, *args, **kwargs)
+
+            return wrapper
+
+        return decorate
+
     def _parameters(self, path):
         """Decorator to handle request parameters."""
 
@@ -1424,7 +1482,7 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    def _perm_auth(self, path, first, *perms):
+    def _perm_authz(self, path, first, *perms):
         """Decorator for per-object permissions."""
         # check perms wrt to recorded per-object checks
 
@@ -1486,20 +1544,6 @@ class FlaskSimpleAuth:
 
         return decorate
 
-    # just to record that no authorization check was needed
-    def _any_noauth(self, path, *groups):
-
-        def decorate(fun: Callable):
-
-            @functools.wraps(fun)
-            def wrapper(*args, **kwargs):
-                self._local.need_authorization = False
-                return self._safe_call(path, "no authorization", fun, *args, **kwargs)
-
-            return wrapper
-
-        return decorate
-
     # FIXME endpoint?
     def add_url_rule(self, rule, endpoint=None, view_func=None, authorize=NONE, auth=None, **options):
         """Route decorator helper method."""
@@ -1542,7 +1586,9 @@ class FlaskSimpleAuth:
             raise self._Bad(f"unexpected authorizations on {rule}: {bads}")
 
         if NONE in predefs:
-            groups, perms = [], []
+            # overwrite all perms, a route is closed just by appending "NONE"
+            # NOTE the handling is performed later to allow for some checks
+            predefs, groups, perms = [NONE], [], []
         elif ANY in predefs:
             if len(predefs) > 1:
                 raise self._Bad(f"cannot mix ANY/ALL predefined groups on {path}")
@@ -1607,27 +1653,32 @@ class FlaskSimpleAuth:
         # else only add needed filters on top of "fun", in reverse order
         need_authenticate = ALL in predefs or groups or perms
         need_parameters = len(fun.__code__.co_varnames) > 0
+        assert len(predefs) <= 1
 
+        # build handling layers in reverse order:
+        # authenticate / (oauth|group|no|) / params / perms / fun
         if perms:
             if not need_parameters:
                 raise self._Bad("permissions require some parameters")
             assert need_authenticate and need_parameters
             first = fun.__code__.co_varnames[0]
-            fun = self._perm_auth(newpath, first, *perms)(fun)
+            fun = self._perm_authz(newpath, first, *perms)(fun)
         if need_parameters:
             fun = self._parameters(newpath)(fun)
         if groups:
             assert need_authenticate
             if auth == "oauth":
-                fun = self._oauth_auth(newpath, *groups)(fun)
+                fun = self._oauth_authz(newpath, *groups)(fun)
             else:
-                fun = self._group_auth(newpath, *groups)(fun)
-        if ANY in predefs:
+                fun = self._group_authz(newpath, *groups)(fun)
+        elif ANY in predefs:
             assert not groups and not perms
-            fun = self._any_noauth(newpath, *groups)(fun)
-        if ALL in predefs:
+            fun = self._no_authz(newpath, *groups)(fun)
+        elif ALL in predefs:
             assert need_authenticate
-            fun = self._any_noauth(newpath, *groups)(fun)
+            fun = self._no_authz(newpath, *groups)(fun)
+        else:  # no authorization at this level
+            assert perms
         if need_authenticate:
             assert perms or groups or ALL in predefs
             fun = self._authenticate(newpath, auth=auth)(fun)
