@@ -15,7 +15,8 @@ log = logging.getLogger("app")
 import FlaskSimpleAuth as fsa
 from FlaskSimpleAuth import Flask, jsonify, ANY, ALL, NONE, path, string, \
     Request, Session, Globals, Environ, CurrentUser
-app = Flask("Test", debug=True)
+app = Flask("Test", debug=True, FSA_LOGGING_LEVEL=logging.DEBUG)
+app.add_group(ADMIN, WRITE, READ)
 
 #
 # AUTH
@@ -24,7 +25,6 @@ app.config.update(
     FSA_AUTH = "fake",
     FSA_GET_USER_PASS = get_user_pass,
     FSA_USER_IN_GROUP = user_in_group,
-    FSA_LOGGING_LEVEL = logging.DEBUG,
 )
 
 
