@@ -497,7 +497,7 @@ def test_password_quality():
 def test_authorize():
     assert app._fsa._user_in_group("dad", App.ADMIN)
     assert not app._fsa._user_in_group("hobbes", App.ADMIN)
-    @app._fsa._group_auth("stuff", App.ADMIN)
+    @app._fsa._group_authz("stuff", App.ADMIN)
     def stuff():
         return Response("", 200)
     app._fsa._local.user = "dad"
@@ -507,7 +507,7 @@ def test_authorize():
     res = stuff()
     assert res.status_code == 403
     try:
-        @app._fsa._group_auth("stuff", fsa.ALL, fsa.ANY)
+        @app._fsa._group_authz("stuff", fsa.ALL, fsa.ANY)
         def foo():
             return "foo", 200
         assert False, "cannot mix ALL & ANY in authorize"
@@ -1218,6 +1218,24 @@ def test_authorize_errors():
         assert False, "should detect ANY/other mix"
     except ConfigError as e:
         assert "object" in str(e)
+    try:
+        app.add_group("foo", "bla")
+        @app.get("/bad-group", authorize="no-such-group")
+        def get_bad_group():
+            return "should not get there", 200
+        assert False, "should detect unregistered group"
+    except ConfigError as e:
+        assert "no-such-group" in str(e)
+    try:
+        app.add_scope("foo", "bla")
+        app._fsa._token = "jwt"
+        app._fsa._issuer = "calvin"
+        @app.get("/bad-scope", authorize="no-such-scope", auth="oauth")
+        def get_bad_scope():
+            return "should not get there", 200
+        assert False, "should detect unregistered scope"
+    except ConfigError as e:
+        assert "no-such-scope" in str(e)
 
 def test_group_errors():
     import AppFact as af
